import copy
import ipaddress

from celery import Celery
from celery.signals import worker_process_init
import jinja2
import keystoneauth1
import openstack
from pottery import Redlock
from redis import Redis

from osism.tasks import Config, conductor, netbox
from osism import utils

app = Celery("openstack")
app.config_from_object(Config)

redis = None
conn = None


@worker_process_init.connect
def celery_init_worker(**kwargs):
    global conn
    global redis

    redis = Redis(host="redis", port="6379")

    # Parameters come from the environment, OS_*
    try:
        conn = openstack.connect()
    except keystoneauth1.exceptions.auth_plugins.MissingRequiredOptions:
        pass


@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    pass


@app.task(bind=True, name="osism.tasks.openstack.image_get")
def image_get(self, image_name):
    result = conn.image.find_image(image_name)
    return result.id


@app.task(bind=True, name="osism.tasks.openstack.network_get")
def network_get(self, network_name):
    result = conn.network.find_network(network_name)
    return result.id


@app.task(bind=True, name="osism.tasks.openstack.baremetal_node_show")
def baremetal_node_show(self, node_id_or_name):
    result = conn.baremetal.find_node(node_id_or_name)
    return result


@app.task(bind=True, name="osism.tasks.openstack.baremetal_node_list")
def baremetal_node_list(self):
    nodes = conn.baremetal.nodes()
    result = []

    # Simulate the output of the OpenStack CLI with -f json and without --long
    for node in nodes:
        result.append(
            {
                "UUID": node.id,
                "Name": node.name,
                "Instance UUID": node.instance_id,
                "Power State": node.power_state,
                "Provisioning State": node.provision_state,
                "Maintenance": node.is_maintenance,
            }
        )

    return result


@app.task(
    bind=True, name="osism.tasks.openstack.baremetal_introspection_interface_list"
)
def baremetal_introspection_interface_list(self, node_id_or_name):
    pass


@app.task(bind=True, name="osism.tasks.openstack.baremetal_introspection_status")
def baremetal_introspection_status(self, node_id_or_name):
    result = None
    return result


@app.task(bind=True, name="osism.tasks.openstack.baremetal_get_network_interface_name")
def baremetal_get_network_interface_name(self, node_name, mac_address):
    global conn

    introspection = conn.baremetal_introspection.get_introspection(node_name)

    # Wait up to 5 minutes for the completion of a running introspection
    conn.baremetal_introspection.wait_for_introspection(introspection, timeout=30)

    introspection_data = conn.baremetal_introspection.get_introspection_data(
        introspection
    )
    interfaces = introspection_data["inventory"]["interfaces"]

    result = None
    for interface in interfaces:
        if interface["mac_address"].lower() == mac_address.lower():
            result = interface["name"]

    return result


@app.task(bind=True, name="osism.tasks.openstack.baremetal_set_node_provision_state")
def baremetal_set_node_provision_state(self, node, state):
    global conn
    conn.baremetal.set_node_provision_state(node, state)


@app.task(bind=True, name="osism.tasks.openstack.baremetal_create_allocations")
def baremetal_create_allocations(self, nodes):
    global conn

    for node in nodes:
        try:
            allocation_a = conn.baremetal.get_allocation(allocation=node)
        except openstack.exceptions.ResourceNotFound:
            allocation_a = None

        if not allocation_a:
            # Get Ironic parameters from the conductor
            task = conductor.get_ironic_parameters.delay()
            task.wait(timeout=None, interval=0.5)
            ironic_parameters = task.get()

            allocation_a = conn.baremetal.create_allocation(
                name=node,
                candidate_nodes=[node],
                resource_class=ironic_parameters["resource_class"],
            )
            conn.baremetal.wait_for_allocation(allocation=node, timeout=30)


@app.task(bind=True, name="osism.tasks.openstack.baremetal_create_nodes")
def baremetal_create_nodes(self, nodes, ironic_parameters):
    global conn

    for node in nodes:
        # TODO: Filter on mgmt_only
        address_a = utils.nb.ipam.ip_addresses.get(device=node, interface="Ethernet0")

        node_parameters = copy.deepcopy(ironic_parameters)

        if node_parameters["driver"] == "redfish":
            remote_board_address = str(ipaddress.ip_interface(address_a["address"]).ip)
            t = jinja2.Environment(loader=jinja2.BaseLoader()).from_string(
                node_parameters["driver_info"]["redfish_address"]
            )
            node_parameters["driver_info"]["redfish_address"] = t.render(
                remote_board_address=remote_board_address
            )

        elif node_parameters["driver"] == "ipmi":
            remote_board_address = str(ipaddress.ip_interface(address_a["address"]).ip)
            t = jinja2.Environment(loader=jinja2.BaseLoader()).from_string(
                node_parameters["driver_info"]["ipmi_address"]
            )
            node_parameters["driver_info"]["ipmi_address"] = t.render(
                remote_board_address=remote_board_address
            )

        try:
            device_a = utils.nb.dcim.devices.get(name=node)
            tags = [str(tag) for tag in device_a.tags]

            # NOTE: Internally used nodes are identified by their unique name via the resource class.
            #       The actual resource class is explicitly overwritten.
            if "Managed by Ironic" in tags and "Managed by OSISM" in tags:
                node_parameters["resource_class"] = f"osism-{node}"
                baremetal_create_internal_flavor(node)

            conn.baremetal.create_node(
                name=node, provision_state="manageable", **node_parameters
            )
            conn.baremetal.wait_for_nodes_provision_state([node], "manageable")

            if "Managed by Ironic" in tags and "Managed by OSISM" not in tags:
                conn.baremetal.set_node_traits(node, ["CUSTOM_GENERAL_USE"])
            elif "Managed by Ironic" in tags and "Managed by OSISM" in tags:
                conn.baremetal.set_node_traits(node, ["CUSTOM_OSISM_USE"])

            conn.baremetal.set_node_provision_state(node, "inspect")

            # TODO: Check if the system has been registered correctly
            device_a.custom_fields = {
                "ironic_state": "registered",
            }
            device_a.save()

        except openstack.exceptions.ResourceFailure:
            # TODO: Do something useful here
            pass
        except openstack.exceptions.ConflictException:
            # The node already exists and has a wronge state in the Netbox
            device_a = utils.nb.dcim.devices.get(name=node)
            device_a.custom_fields = {
                "ironic_state": "registered",
            }
            device_a.save()


@app.task(bind=True, name="osism.tasks.openstack.baremetal_check_allocations")
def baremetal_check_allocations(self):
    lock = Redlock(
        key="lock_osism_tasks_openstack_baremetal_check_allocations",
        masters={redis},
        auto_release_time=60,
    )

    if lock.acquire(timeout=20):
        netbox.get_devices_that_should_have_an_allocation_in_ironic.apply_async(
            (), link=baremetal_create_allocations.s()
        )
        lock.release()


@app.task(bind=True, name="osism.tasks.openstack.baremetal_create_internal_flavor")
def baremetal_create_internal_flavor(self, node):
    flavor_a = conn.compute.create_flavor(
        name=f"osism-{node}", ram=1, vcpus=1, disk=1, is_public=False
    )
    specs = {
        f"resources:CUSTOM_RESOURCE_CLASS_OSISM_{node.upper()}": 1,
        "resources:VCPU": 0,
        "resources:MEMORY_MB": 0,
        "resources:DISK_GB": 0,
        "trait:CUSTOM_OSISM_USE": "required",
    }
    conn.compute.create_flavor_extra_specs(flavor_a, specs)


@app.task(bind=True, name="osism.tasks.openstack.baremetal_delete_internal_flavor")
def baremetal_delete_internal_flavor(self, node):
    flavor = conn.compute.get_flavor(f"osism-{node}")
    conn.compute.delete_flavor(flavor)



import io
import os
from pathlib import Path
import subprocess

# import ansible_runner
from celery.signals import worker_process_init
from redis import Redis
from pottery import Redlock


redis = None


class Config:
    enable_utc = True
    enable_bifrost = os.environ.get("ENABLE_BIFROST", "False")
    enable_ironic = os.environ.get("ENABLE_IRONIC", "True")
    broker_url = "redis://redis"
    result_backend = "redis://redis"
    task_create_missing_queues = True
    task_default_queue = "default"
    task_track_started = (True,)
    task_routes = {
        "osism.tasks.ceph.*": {"queue": "ceph-ansible"},
        "osism.tasks.conductor.*": {"queue": "conductor"},
        "osism.tasks.kolla.*": {"queue": "kolla-ansible"},
        "osism.tasks.netbox.*": {"queue": "netbox"},
        "osism.tasks.ansible.*": {"queue": "osism-ansible"},
        "osism.tasks.reconciler.*": {"queue": "reconciler"},
        "osism.tasks.openstack.*": {"queue": "openstack"},
    }


@worker_process_init.connect
def celery_init_worker(**kwargs):
    global redis

    redis = Redis(host="redis", port="6379")


def run_ansible_in_environment(request_id, environment, role, arguments, publish=True):
    result = ""

    if type(arguments) == list:
        joined_arguments = " ".join(arguments)
    else:
        joined_arguments = arguments

    # NOTE: Consider arguments in the future
    lock = Redlock(
        key=f"lock-ansible-{environment}-{role}",
        masters={redis},
        auto_release_time=3600,
    )

    # NOTE: use python interface in the future, something with ansible-runner and the fact cache is
    #       not working out of the box

    # execute roles from kolla-ansible
    if environment == "kolla" and not role.startswith("validate"):
        lock.acquire()

        if role == "mariadb_backup":
            p = subprocess.Popen(
                f"/run.sh backup {role} {joined_arguments}",
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
        else:
            p = subprocess.Popen(
                f"/run.sh deploy {role} {joined_arguments}",
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )

    # execute roles from ceph-ansible
    elif environment == "ceph" and not role.startswith("validate"):
        lock.acquire()
        p = subprocess.Popen(
            f"/run.sh {role} {joined_arguments}",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

    # execute the bifrost-command role
    elif environment == "manager" and role == "bifrost-command":
        p = subprocess.Popen(
            f'/run-manager.sh bifrost-command "-e bifrost_arguments=\'{joined_arguments}\'" "-e bifrost_result_id={request_id}"',
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

    # execute all other roles
    else:
        lock.acquire()
        p = subprocess.Popen(
            f"/run-{environment}.sh {role} {joined_arguments}",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

    # process the bifrost-command result
    if environment == "manager" and role == "bifrost-command":
        p.wait()

        # Check for JSON result
        resultpath = f"/tmp/bifrost-command-{request_id}.json"
        if os.path.exists(resultpath):
            result = Path(resultpath).read_text()
            os.remove(resultpath)

        # Check for non-JSON result
        resultpath = f"/tmp/bifrost-command-{request_id}.log"
        if os.path.exists(resultpath):
            result = Path(resultpath).read_text()
            os.remove(resultpath)

    # process all other results
    else:
        for line in io.TextIOWrapper(p.stdout, encoding="utf-8"):
            if publish:
                redis.publish(f"{request_id}", line)
            result += line

        rc = p.wait(timeout=60)

        if publish:
            redis.publish(f"{request_id}", f"RC: {rc}\n")
            redis.publish(f"{request_id}", "QUIT")

        lock.release()

    return result


""" def run_ansible_in_environment(request_id, environment, playbook, arguments):
    os.mkdir(f"/tmp/{request_id}")

    # NOTE: check for existence of ansible.cfg inside the used environment

    # NOTE: https://docs.ansible.com/ansible/latest/reference_appendices/config.html
    envvars = {
        "ANSIBLE_CONFIG": "/opt/configuration/environments/ansible.cfg",
        "ANSIBLE_DIRECTORY": "/ansible",
        "ANSIBLE_INVENTORY": "/ansible/inventory",
        "CACHE_PLUGIN": "redis",
        "CACHE_PLUGIN_CONNECTION": "cache:6379:0",
        "CACHE_PLUGIN_TIMEOUT": "86400",
        "CONFIGURATION_DIRECTORY": "/opt/configuration",
        "ENVIRONMENTS_DIRECTORY": "/opt/configuration/environments",
        "ANSIBLE_CALLBACK_PLUGINS": "/usr/local/lib/python3.8/dist-packages/ara/plugins/callback",
        "ANSIBLE_ACTION_PLUGINS": "/usr/local/lib/python3.8/dist-packages/ara/plugins/action",
        "ANSIBLE_LOOKUP_PLUGINS": "/usr/local/lib/python3.8/dist-packages/ara/plugins/lookup"
    }

    extravars = {}

    # NOTE: https://ansible-runner.readthedocs.io/en/stable/intro.html#env-settings-settings-for-runner-itself
    settings = {
        "fact_cache_type": "json"
    }

    cmdline = [
        "--vault-password-file /opt/configuration/environments/.vault_pass",
        f"-e @/opt/configuration/environments/{environment}/configuration.yml",
        f"-e @/opt/configuration/environments/{environment}/secrets.yml",
        "-e @secrets.yml",
        "-e @images.yml",
        "-e @configuration.yml"
    ]

    if environment == "kolla":
        extravars["CONFIG_DIR"] = f"/opt/configuration/environments/{environment}"
        extravars["kolla_action"] = "deploy"

    if environment == "ceph":
        cmdline.append("--skip-tags=with_pkg")

    # NOTE: https://github.com/ansible/ansible-runner/blob/devel/ansible_runner/interface.py
    ansible_runner.interface.run(
        private_data_dir=f"/tmp/{request_id}",
        ident=request_id,
        project_dir="/opt/configuration/environments",
        inventory="/ansible/inventory",
        playbook=f"/ansible/{environment}-{playbook}.yml",
        envvars=envvars,
        extravars=extravars,
        cmdline=" ".join(cmdline + arguments),
        settings=settings
    )
 """



LOADBALANCER_PLAYBOOKS = [
    "loadbalancer-aodh",
    "loadbalancer-barbican",
    # "loadbalancer-blazar",
    # "loadbalancer-ceph-rgw",
    "loadbalancer-cinder",
    # "loadbalancer-cloudkitty",
    # "loadbalancer-cyborg",
    "loadbalancer-designate",
    "loadbalancer-elasticsearch",
    # "loadbalancer-freezer",
    "loadbalancer-glance",
    "loadbalancer-gnocchi",
    "loadbalancer-grafana",
    "loadbalancer-heat",
    "loadbalancer-horizon",
    # "loadbalancer-influxdb",
    "loadbalancer-ironic",
    "loadbalancer-keystone",
    "loadbalancer-kibana",
    # "loadbalancer-magnum",
    "loadbalancer-manila",
    "loadbalancer-mariadb",
    # "loadbalancer-masakari",
    "loadbalancer-memcached",
    # "loadbalancer-mistral",
    # "loadbalancer-monasca",
    # "loadbalancer-murano",
    "loadbalancer-neutron",
    "loadbalancer-nova",
    "loadbalancer-octavia",
    "loadbalancer-placement",
    "loadbalancer-prometheus",
    "loadbalancer-rabbitmq",
    # "loadbalancer-sahara",
    # "loadbalancer-senlin",
    "loadbalancer-skydive",
    # "loadbalancer-solum",
    # "loadbalancer-swift",
    # "loadbalancer-tacker",
    # "loadbalancer-trove",
    # "loadbalancer-vitrage",
    # "loadbalancer-watcher",
    # "loadbalancer-zun",
]

VALIDATE_PLAYBOOK2ENVIRONMENT = {
    "ceph-mgrs": "ceph",
    "ceph-mons": "ceph",
    "ceph-osds": "ceph",
    "container-status": "generic",
    "kernel-version": "generic",
    "mysql-open-files-limit": "generic",
    "system-encoding": "generic",
    "ulimits": "generic",
}

MAP_ROLE2ROLE = {
    "ceph-basic": [
        "ceph-mons",
        "ceph-mgrs",
        "ceph-osds",
        "ceph-crash",
    ],
    "infrastructure-basic": [
        "openstackclient",
        "common",
        "loadbalancer",
        "elasticsearch",
        "kibana",
        "openvswitch",
        "memcached",
        "redis",
        "mariadb",
        "rabbitmq",
        "phpmyadmin",
    ],
    "openstack-basic": [
        "keystone",
        "horizon",
        "placement",
        "glance",
        "cinder",
        "neutron",
        "nova",
        "barbican",
        "designate",
        "heat",
        "octavia",
    ],
    "openstack-extended": ["gnocchi", "ceilometer", "aodh", "senlin"],
}

MAP_ROLE2ENVIRONMENT = {
    # MONITORING
    "netdata": "monitoring",
    "remove-netdata": "monitoring",
    "remove-zabbix-agent": "monitoring",
    "openstack-health-monitor": "monitoring",
    # GENERIC
    "auditd": "generic",
    "backup-mariadb": "generic",
    "bootstrap": "generic",
    "check-reboot": "generic",
    "chrony": "generic",
    "chrony-force-sync": "generic",
    "clamav": "generic",
    "cleanup": "generic",
    "cleanup-backup-mariadb": "generic",
    "cleanup-databases": "generic",
    "cleanup-docker": "generic",
    "cleanup-docker-images": "generic",
    "cleanup-elasticsearch": "generic",
    "cleanup-queues": "generic",
    "cleanup-sosreport": "generic",
    "cockpit": "generic",
    "configfs": "generic",
    "docker": "generic",
    "docker-compose": "generic",
    "dotfiles": "generic",
    "dump-facts": "generic",
    "facts": "generic",
    "fail2ban": "generic",
    "falco": "generic",
    "firewall": "generic",
    "grub": "generic",
    "halt": "generic",
    "hardening": "generic",
    "hddtemp": "generic",
    "hostname": "generic",
    "hosts": "generic",
    "ipmitool": "generic",
    "journald": "generic",
    "kernel-modules": "generic",
    "known-hosts": "generic",
    "kompose": "generic",
    "lldpd": "generic",
    "lynis": "generic",
    "maintenance": "generic",
    "manage-container": "generic",
    "manage-service": "generic",
    "microcode": "generic",
    "motd": "generic",
    "network": "generic",
    "operator": "generic",
    "osquery": "generic",
    "packages": "generic",
    "patchman-client": "generic",
    "ping": "generic",
    "podman": "generic",
    "proxy": "generic",
    "python": "generic",
    "python3": "generic",
    "reboot": "generic",
    "remove-deploy-user": "generic",
    "repository": "generic",
    "resolvconf": "generic",
    "rng": "generic",
    "rsyslog": "generic",
    "services": "generic",
    "smartd": "generic",
    "sosreport": "generic",
    "state": "generic",
    "sysctl": "generic",
    "sysdig": "generic",
    "systohc": "generic",
    "timezone": "generic",
    "trivy": "generic",
    "upgrade-packages": "generic",
    "utilities": "generic",
    "wait-for-connection": "generic",
    "write-facts": "generic",
    # INFRASTRUCTURE
    "adminer": "infrastructure",
    "cephclient": "infrastructure",
    "cgit": "infrastructure",
    "clevis": "infrastructure",
    "dnsdist": "infrastructure",
    "helper": "infrastructure",
    "homer": "infrastructure",
    "jenkins": "infrastructure",
    "keycloak": "infrastructure",
    "kubectl": "infrastructure",
    "minikube": "infrastructure",
    "mirror": "infrastructure",
    "mirror-images": "infrastructure",
    "netbox": "infrastructure",
    "nexus": "infrastructure",
    "openldap": "infrastructure",
    "openstackclient": "infrastructure",
    "patchman": "infrastructure",
    "phpmyadmin": "infrastructure",
    "rundeck": "infrastructure",
    "squid": "infrastructure",
    "sshconfig": "infrastructure",
    "tailscale": "infrastructure",
    "tang": "infrastructure",
    "traefik": "infrastructure",
    "virtualbmc": "infrastructure",
    "wireguard": "infrastructure",
    "zuul": "infrastructure",
    # MANAGER
    "configuration": "manager",
    "copy-ceph-keys": "manager",
    "manager-network": "manager",
    "manager-operator": "manager",
    "vault-import": "manager",
    "vault-init": "manager",
    "vault-seal": "manager",
    "vault-unseal": "manager",
    # CEPH
    "ceph-add-mon": "ceph",
    "ceph-bootstrap-dashboard": "ceph",
    "ceph-ceph-keys": "ceph",
    "ceph-cephadm-adopt": "ceph",
    "ceph-cephadm": "ceph",
    "ceph-clients": "ceph",
    "ceph-config": "ceph",
    "ceph-crash": "ceph",
    "ceph-docker-to-podman": "ceph",
    "ceph-facts": "ceph",
    "ceph-fetch-keys": "ceph",
    "ceph-filestore-to-bluestore": "ceph",
    "ceph-gather-ceph-logs": "ceph",
    "ceph-iscsigws": "ceph",
    "ceph-lv-create": "ceph",
    "ceph-lv-teardown": "ceph",
    "ceph-mdss": "ceph",
    "ceph-mgrs": "ceph",
    "ceph-mons": "ceph",
    "ceph-nfss": "ceph",
    "ceph-osds": "ceph",
    "ceph-purge-cluster": "ceph",
    "ceph-purge-storage-node": "ceph",
    "ceph-rbd-mirrors": "ceph",
    "ceph-restapis": "ceph",
    "ceph-rgw-add-users-buckets": "ceph",
    "ceph-rgws": "ceph",
    "ceph-rolling_update": "ceph",
    "ceph-shrink-mds": "ceph",
    "ceph-shrink-mgr": "ceph",
    "ceph-shrink-mon": "ceph",
    "ceph-shrink-osd": "ceph",
    "ceph-shrink-rbdmirror": "ceph",
    "ceph-shrink-rgw": "ceph",
    "ceph-site": "ceph",
    "ceph-storage-inventory": "ceph",
    "ceph-switch-from-non-containerized-to-containerized-ceph-daemons": "ceph",
    "ceph-take-over-existing-cluster": "ceph",
    "ceph-base": "ceph",
    # KOLLA
    "aodh": "kolla",
    "barbican": "kolla",
    "bifrost-keypair": "kolla",
    "bifrost": "kolla",
    "blazar": "kolla",
    "ceilometer": "kolla",
    "certificates": "kolla",
    "chrony-cleanup": "kolla",
    # "chrony": "kolla",
    "cinder": "kolla",
    "cloudkitty": "kolla",
    "collectd": "kolla",
    "common": "kolla",
    "cyborg": "kolla",
    "designate": "kolla",
    "kolla-destroy": "kolla",
    "elasticsearch": "kolla",
    "etcd": "kolla",
    "kolla-facts": "kolla",
    "freezer": "kolla",
    "kolla-gather-facts": "kolla",
    "glance": "kolla",
    "gnocchi": "kolla",
    "grafana": "kolla",
    "hacluster": "kolla",
    "haproxy": "kolla",
    "heat": "kolla",
    "horizon": "kolla",
    "influxdb": "kolla",
    "ironic": "kolla",
    "iscsi": "kolla",
    "kafka": "kolla",
    "keystone": "kolla",
    "kibana": "kolla",
    "kuryr": "kolla",
    "loadbalancer": "kolla",
    "magnum": "kolla",
    "manila": "kolla",
    "mariadb-dynamic-rows": "kolla",
    "mariadb": "kolla",
    "mariadb_backup": "kolla",
    "mariadb_recovery": "kolla",
    "masakari": "kolla",
    "memcached": "kolla",
    "mistral": "kolla",
    "monasca": "kolla",
    "monasca_cleanup": "kolla",
    "multipathd": "kolla",
    "murano": "kolla",
    "neutron": "kolla",
    "nova-compute": "kolla",
    "nova": "kolla",
    "octavia-certificates": "kolla",
    "octavia": "kolla",
    "openvswitch": "kolla",
    "ovn": "kolla",
    "ovn-controller": "kolla",
    "ovn-db": "kolla",
    "ovs-dpdk": "kolla",
    "panko": "kolla",
    "placement": "kolla",
    "prechecks": "kolla",
    "prometheus": "kolla",
    "kolla-prune-images": "kolla",
    "kolla-purge": "kolla",
    "qdrouterd": "kolla",
    "rabbitmq-outward": "kolla",
    "rabbitmq": "kolla",
    "rally": "kolla",
    "redis": "kolla",
    "kolla-rgw-endpoint": "kolla",
    "sahara": "kolla",
    "senlin": "kolla",
    "kolla-site": "kolla",
    "skydive": "kolla",
    "solum": "kolla",
    "storm": "kolla",
    "swift": "kolla",
    "tacker": "kolla",
    "telegraf": "kolla",
    "tempest": "kolla",
    "kolla-testbed-identity": "kolla",
    "kolla-testbed": "kolla",
    "trove": "kolla",
    "vitrage": "kolla",
    "vmtp": "kolla",
    "watcher": "kolla",
    "zookeeper": "kolla",
    "zun": "kolla",
}



#!/usr/bin/env python3

from collections import Counter
from datetime import datetime
import argparse
import glob
import os

import pynetbox
import yaml

from osism import settings


counter = Counter(added=0, updated=0, manufacturer=0)


def slugFormat(name):
    return name.lower().replace(' ', '_')

YAML_EXTENSIONS = ['yml', 'yaml']

def getFiles(vendors=None):
    files = []
    discoveredVendors = []
    base_path = settings.BASE_PATH
    if vendors:
        for r, d, f in os.walk(base_path):
            for folder in d:
                for vendor in vendors:
                    if vendor.lower() == folder.lower():
                        discoveredVendors.append({'name': folder,
                                                  'slug': slugFormat(folder)})
                        for extension in YAML_EXTENSIONS:
                            files.extend(glob.glob(base_path + folder + f'/*.{extension}'))
    else:
        for r, d, f in os.walk(base_path):
            for folder in d:
                if folder.lower() != "Testing":
                    discoveredVendors.append({'name': folder,
                                              'slug': slugFormat(folder)})
        for extension in YAML_EXTENSIONS:
            files.extend(glob.glob(base_path + f'[!Testing]*/*.{extension}'))
    return files, discoveredVendors


def readYAMl(files):
    deviceTypes = []
    manufacturers = []
    for file in files:
        with open(file, 'r') as stream:
            try:
                data = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)
                continue
            manufacturer = data['manufacturer']
            data['manufacturer'] = {}
            data['manufacturer']['name'] = manufacturer
            data['manufacturer']['slug'] = slugFormat(manufacturer)
        deviceTypes.append(data)
        manufacturers.append(manufacturer)
    return deviceTypes


def createManufacturers(vendors, nb):
    all_manufacturers = {str(item): item for item in nb.dcim.manufacturers.all()}
    need_manufacturers = []
    for vendor in vendors:
        try:
            manGet = all_manufacturers[vendor["name"]]
            print(f'Manufacturer Exists: {manGet.name} - {manGet.id}')
        except KeyError:
            need_manufacturers.append(vendor)

    if not need_manufacturers:
        return

    try:
        manSuccess = nb.dcim.manufacturers.create(need_manufacturers)
        for man in manSuccess:
            print(f'Manufacturer Created: {man.name} - '
                  + f'{man.id}')
            counter.update({'manufacturer': 1})
    except pynetbox.RequestError as e:
        print(e.error)


def createInterfaces(interfaces, deviceType, nb):
    all_interfaces = {str(item): item for item in nb.dcim.interface_templates.filter(devicetype_id=deviceType)}
    need_interfaces = []
    for interface in interfaces:
        try:
            ifGet = all_interfaces[interface["name"]]
            print(f'Interface Template Exists: {ifGet.name} - {ifGet.type}'
                  + f' - {ifGet.device_type.id} - {ifGet.id}')
        except KeyError:
            interface['device_type'] = deviceType
            need_interfaces.append(interface)

    if not need_interfaces:
        return
    
    try:
        ifSuccess = nb.dcim.interface_templates.create(need_interfaces)
        for intf in ifSuccess:
            print(f'Interface Template Created: {intf.name} - '
              + f'{intf.type} - {intf.device_type.id} - '
              + f'{intf.id}')
            counter.update({'updated': 1})
    except pynetbox.RequestError as e:
        print(e.error)
    


def createConsolePorts(consoleports, deviceType, nb):
    all_consoleports = {str(item): item for item in nb.dcim.console_port_templates.filter(devicetype_id=deviceType)}
    need_consoleports = []
    for consoleport in consoleports:
        try:
            cpGet = all_consoleports[consoleport["name"]]
            print(f'Console Port Template Exists: {cpGet.name} - '
                  + f'{cpGet.type} - {cpGet.device_type.id} - {cpGet.id}')
        except KeyError:
            consoleport['device_type'] = deviceType
            need_consoleports.append(consoleport)

    if not need_consoleports:
        return
                
    try:
        cpSuccess = nb.dcim.console_port_templates.create(need_consoleports)
        for port in cpSuccess:
            print(f'Console Port Created: {port.name} - '
                  + f'{port.type} - {port.device_type.id} - '
                  + f'{port.id}')
            counter.update({'updated': 1})
    except pynetbox.RequestError as e:
        print(e.error)


def createPowerPorts(powerports, deviceType, nb):
    all_power_ports = {str(item): item for item in nb.dcim.power_port_templates.filter(devicetype_id=deviceType)}
    need_power_ports = []
    for powerport in powerports:
        try:
            ppGet = all_power_ports[powerport["name"]]
            print(f'Power Port Template Exists: {ppGet.name} - '
                  + f'{ppGet.type} - {ppGet.device_type.id} - {ppGet.id}')            
        except KeyError:
            powerport['device_type'] = deviceType
            need_power_ports.append(powerport)

    if not need_power_ports:
        return

    try:
        ppSuccess = nb.dcim.power_port_templates.create(need_power_ports)
        for pp in ppSuccess:
            print(f'Interface Template Created: {pp.name} - '
              + f'{pp.type} - {pp.device_type.id} - '
              + f'{pp.id}')
            counter.update({'updated': 1})
    except pynetbox.RequestError as e:
        print(e.error)


def createConsoleServerPorts(consoleserverports, deviceType, nb):
    all_consoleserverports = {str(item): item for item in nb.dcim.console_server_port_templates.filter(devicetype_id=deviceType)}
    need_consoleserverports = []
    for csport in consoleserverports:
        try:
            cspGet = all_consoleserverports[csport["name"]]
            print(f'Console Server Port Template Exists: {cspGet.name} - '
                  + f'{cspGet.type} - {cspGet.device_type.id} - '
                  + f'{cspGet.id}')
        except KeyError:
            csport['device_type'] = deviceType
            need_consoleserverports.append(csport)

    if not need_consoleserverports:
        return

    try:
        cspSuccess = nb.dcim.console_server_port_templates.create(
            need_consoleserverports)
        for csp in cspSuccess:
            print(f'Console Server Port Created: {csp.name} - '
                  + f'{csp.type} - {csp.device_type.id} - '
                  + f'{csp.id}')
            counter.update({'updated': 1})
    except pynetbox.RequestError as e:
        print(e.error)


def createFrontPorts(frontports, deviceType, nb):
    all_frontports = {str(item): item for item in nb.dcim.front_port_templates.filter(devicetype_id=deviceType)}
    need_frontports = []
    for frontport in frontports:
        try:
            fpGet = all_frontports[frontport["name"]]
            print(f'Front Port Template Exists: {fpGet.name} - '
                  + f'{fpGet.type} - {fpGet.device_type.id} - {fpGet.id}')
        except KeyError:
            frontport['device_type'] = deviceType
            need_frontports.append(frontport)

    if not need_frontports:
        return

    all_rearports = {str(item): item for item in nb.dcim.rear_port_templates.filter(devicetype_id=deviceType)}
    for port in need_frontports:
        try:
            rpGet = all_rearports[port["rear_port"]]
            port['rear_port'] = rpGet.id
        except KeyError:
            print(f'Could not find Rear Port for Front Port: {port["name"]} - '
                  + f'{port["type"]} - {deviceType}')

    try:
        fpSuccess = nb.dcim.front_port_templates.create(need_frontports)
        for fp in fpSuccess:
            print(f'Front Port Created: {fp.name} - '
                  + f'{fp.type} - {fp.device_type.id} - '
                  + f'{fp.id}')
            counter.update({'updated': 1})
    except pynetbox.RequestError as e:
        print(e.error)


def createRearPorts(rearports, deviceType, nb):
    all_rearports = {str(item): item for item in nb.dcim.rear_port_templates.filter(devicetype_id=deviceType)}
    need_rearports = []
    for rearport in rearports:
        try:
            rpGet = all_rearports[rearport["name"]]
            print(f'Rear Port Template Exists: {rpGet.name} - {rpGet.type}'
                  + f' - {rpGet.device_type.id} - {rpGet.id}')
        except KeyError:
            rearport['device_type'] = deviceType
            need_rearports.append(rearport)

    if not need_rearports:
        return

    try:
        rpSuccess = nb.dcim.rear_port_templates.create(
            need_rearports)
        for rp in rpSuccess:
            print(f'Rear Port Created: {rp.name} - {rp.type}'
                  + f' - {rp.device_type.id} - {rp.id}')
            counter.update({'updated': 1})
    except pynetbox.RequestError as e:
        print(e.error)


def createDeviceBays(devicebays, deviceType, nb):
    all_devicebays = {str(item): item for item in nb.dcim.device_bay_templates.filter(devicetype_id=deviceType)}
    need_devicebays = []
    for devicebay in devicebays:
        try:
            dbGet = all_devicebays[devicebay["name"]]
            print(f'Device Bay Template Exists: {dbGet.name} - '
                  + f'{dbGet.device_type.id} - {dbGet.id}')
        except KeyError:
            devicebay['device_type'] = deviceType
            need_devicebays.append(devicebay)

    if not need_devicebays:
        return

    try:
        dbSuccess = nb.dcim.device_bay_templates.create(need_devicebays)
        for db in dbSuccess:
            print(f'Device Bay Created: {db.name} - '
                  + f'{db.device_type.id} - {db.id}')
            counter.update({'updated': 1})
    except pynetbox.RequestError as e:
        print(e.error)


def createPowerOutlets(poweroutlets, deviceType, nb):
    all_poweroutlets = {str(item): item for item in nb.dcim.power_outlet_templates.filter(devicetype_id=deviceType)}
    need_poweroutlets = []
    for poweroutlet in poweroutlets:
        try:
            poGet = all_poweroutlets[poweroutlet["name"]]
            print(f'Power Outlet Template Exists: {poGet.name} - '
                  + f'{poGet.type} - {poGet.device_type.id} - {poGet.id}')
        except KeyError:
            poweroutlet["device_type"] = deviceType
            need_poweroutlets.append(poweroutlet)

    if not need_poweroutlets:
        return

    all_power_ports = {str(item): item for item in nb.dcim.power_port_templates.filter(devicetype_id=deviceType)}
    for outlet in need_poweroutlets:
        try:
            ppGet = all_power_ports[outlet["power_port"]]
            outlet['power_port'] = ppGet.id
        except KeyError:
            pass

    try:
        poSuccess = nb.dcim.power_outlet_templates.create(
            need_poweroutlets)
        for po in poSuccess:
            print(f'Power Outlet Created: {po.name} - '
                  + f'{po.type} - {po.device_type.id} - '
                  + f'{po.id}')
            counter.update({'updated': 1})
    except pynetbox.RequestError as e:
        print(e.error)


def createDeviceTypes(deviceTypes, nb):
    all_device_types = {str(item): item for item in nb.dcim.device_types.all()}
    for deviceType in deviceTypes:
        try:
            dt = all_device_types[deviceType["model"]]
            print(f'Device Type Exists: {dt.manufacturer.name} - '
                  + f'{dt.model} - {dt.id}')
        except KeyError:
            try:
                dt = nb.dcim.device_types.create(deviceType)
                counter.update({'added': 1})
                print(f'Device Type Created: {dt.manufacturer.name} - '
                      + f'{dt.model} - {dt.id}')
            except pynetbox.RequestError as e:
                print(e.error)

        if "interfaces" in deviceType:
            createInterfaces(deviceType["interfaces"],
                             dt.id, nb)
        if "power-ports" in deviceType:
            createPowerPorts(deviceType["power-ports"],
                             dt.id, nb)
        if "power-port" in deviceType:
            createPowerPorts(deviceType["power-port"],
                             dt.id, nb)
        if "console-ports" in deviceType:
            createConsolePorts(deviceType["console-ports"],
                               dt.id, nb)
        if "power-outlets" in deviceType:
            createPowerOutlets(deviceType["power-outlets"],
                               dt.id, nb)
        if "console-server-ports" in deviceType:
            createConsoleServerPorts(
                deviceType["console-server-ports"], dt.id, nb)
        if "rear-ports" in deviceType:
            createRearPorts(deviceType["rear-ports"],
                            dt.id, nb)
        if "front-ports" in deviceType:
            createFrontPorts(deviceType["front-ports"],
                             dt.id, nb)
        if "device-bays" in deviceType:
            createDeviceBays(deviceType["device-bays"],
                             dt.id, nb)


def main():

    cwd = os.getcwd()
    startTime = datetime.now()

    nbUrl = settings.NETBOX_URL
    nbToken = settings.NETBOX_TOKEN
    nb = pynetbox.api(nbUrl, token=nbToken)

    if settings.IGNORE_SSL_ERRORS:
        import requests
        requests.packages.urllib3.disable_warnings()
        session = requests.Session()
        session.verify = False
        nb.http_session = session


    VENDORS = settings.VENDORS

    parser = argparse.ArgumentParser(description='Import Netbox Device Types')
    parser.add_argument('--vendors', nargs='+', default=VENDORS,
                        help="List of vendors to import eg. apc cisco")
    args = parser.parse_args()

    if not args.vendors:
        print("No Vendors Specified, Gathering All Device-Types")
        files, vendors = getFiles()
    else:
        print("Vendor Specified, Gathering All Matching Device-Types")
        files, vendors = getFiles(args.vendors)


    print(str(len(vendors)) + " Vendors Found")
    print(str(len(files)) + " Device-Types Found")
    deviceTypes = readYAMl(files)
    createManufacturers(vendors, nb)
    createDeviceTypes(deviceTypes, nb)

    print('---')
    print('Script took {} to run'.format(datetime.now() - startTime))
    print('{} devices created'.format(counter['added']))
    print('{} interfaces/ports updated'.format(counter['updated']))
    print('{} manufacturers created'.format(counter['manufacturer']))


if __name__ == "__main__":
    main()



import time

from celery import Celery
from celery.result import AsyncResult
from cliff.command import Command
from loguru import logger
from redis import Redis

from osism.tasks import Config


class Run(Command):
    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument(
            "--delay",
            default=1,
            type=int,
            help="Delay in seconds between two task checks",
        )
        parser.add_argument(
            "--live",
            default=False,
            help="Show live output from a started task until it is finished",
            action="store_true",
        )
        parser.add_argument(
            "--format",
            default="log",
            help="Output type",
            const="log",
            nargs="?",
            choices=["script", "log"],
        ),
        parser.add_argument(
            "--output",
            default=False,
            help="Show output from a finished task",
            action="store_true",
        )
        parser.add_argument(
            "task_id", nargs="+", type=str, help="ID of tasks to wait for"
        )
        return parser

    def take_action(self, parsed_args):
        delay = parsed_args.delay
        format = parsed_args.format
        live = parsed_args.live
        output = parsed_args.output
        task_ids = parsed_args.task_id

        app = Celery("wait")
        app.config_from_object(Config)
        i = app.control.inspect()

        while task_ids:
            time.sleep(delay)

            task_id = task_ids.pop()
            result = AsyncResult(f"{task_id}", app=app)

            if result.state == "PENDING":

                q = i.query_task(f"{task_id}")
                if not len([x for x in q.values() if len(x)]):
                    if format == "log":
                        logger.info(f"Task {task_id} is unavailable")
                    elif format == "script":
                        print(f"{task_id} = UNAVAILABLE")
                else:
                    if format == "log":
                        logger.info(f"Task {task_id} is in state PENDING")
                    elif format == "script":
                        print(f"{task_id} = PENDING")

                    task_ids.insert(0, task_id)

            elif result.state == "SUCCESS":

                if format == "log":
                    logger.info(f"Task {task_id} is in state SUCCESS")
                elif format == "script":
                    print(f"{task_id} = SUCCESS")

                if output:
                    print(result.get())

            elif result.state == "STARTED":

                if format == "log":
                    logger.info(f"Task {task_id} is in state STARTED")
                elif format == "script":
                    print(f"{task_id} = STARTED")

                if live:
                    redis = Redis(host="redis", port="6379")
                    p = redis.pubsub()
                    p.subscribe(f"{task_id}")

                    while_True = True
                    while while_True:
                        for m in p.listen():
                            if type(m["data"]) == bytes:
                                line = m["data"].decode("utf-8")
                                if line.startswith("RC: "):
                                    rc = int(line[4:])
                                    continue
                                elif line == "QUIT":
                                    redis.close()

                                    if len(task_ids) == 1:
                                        return rc
                                    else:
                                        while_True = False
                                else:
                                    print(line, end="")
                else:
                    task_ids.insert(0, task_id)



import io
import subprocess

from celery import Celery
from celery.signals import worker_process_init
from pottery import Redlock
from redis import Redis

from osism.tasks import Config

app = Celery("reconciler")
app.config_from_object(Config)

redis = None


@worker_process_init.connect
def celery_init_worker(**kwargs):
    global redis

    redis = Redis(host="redis", port="6379")


@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    sender.add_periodic_task(600.0, run.s(), expires=10)


@app.task(bind=True, name="osism.tasks.reconciler.run")
def run(self):
    lock = Redlock(
        key="lock_osism_tasks_reconciler_run", masters={redis}, auto_release_time=60
    )

    if lock.acquire(timeout=20):
        p = subprocess.Popen(
            "/run.sh", shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        p.wait()

        lock.release()


@app.task(bind=True, name="osism.tasks.reconciler.sync_inventory_with_netbox")
def sync_inventory_with_netbox(self):
    lock = Redlock(
        key="lock_osism_tasks_reconciler_sync_inventory_with_netbox",
        masters={redis},
        auto_release_time=60,
    )

    if lock.acquire(timeout=20):
        p = subprocess.Popen(
            "/sync-inventory-with-netbox.sh",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        for line in io.TextIOWrapper(p.stdout, encoding="utf-8"):
            # NOTE: use task_id or request_id in future
            redis.publish("netbox-sync-inventory-with-netbox", line)

        lock.release()

    # NOTE: use task_id or request_id in future
    redis.publish("netbox-sync-inventory-with-netbox", "QUIT")



import pprint
import subprocess

from cliff.command import Command
import json
from loguru import logger
from tabulate import tabulate

from osism.utils import redis


class Facts(Command):
    def get_parser(self, prog_name):
        parser = super(Facts, self).get_parser(prog_name)
        parser.add_argument(
            "host",
            nargs=1,
            type=str,
            help="Hostname (as the host is known in Ansible inventory)",
        )
        parser.add_argument(
            "fact",
            nargs="?",
            type=str,
            help="Name of a fact to show",
        )
        parser.add_argument(
            "--no-cache",
            default=False,
            help="Do not use facts from the cache",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        host = parsed_args.host[0]
        fact = parsed_args.fact
        cache = not parsed_args.no_cache

        data = redis.get(f"ansible_facts{host}")
        if data:
            data = json.loads(data)
            table = []

            if fact:
                if fact in data:
                    row = pprint.pformat(data[fact], indent=2, width=60, compact=True)
                    table.append([host, fact, row])
                else:
                    logger.error(f"Fact {fact} not found in cache for {host}.")
            else:
                for fact in data:
                    row = pprint.pformat(data[fact], indent=2, width=60, compact=True)
                    if fact in [
                        "ansible_ssh_host_key_dsa_public",
                        "ansible_ssh_host_key_ecdsa_public",
                        "ansible_ssh_host_key_ed25519_public",
                        "ansible_ssh_host_key_rsa_public",
                    ]:
                        row = f"{row[0:40]}..."
                    table.append([host, fact, row])

            if table:
                print(
                    tabulate(table, headers=["Host", "Fact", "Value"], tablefmt="grid")
                )
        else:
            logger.error(f"No facts found in cache for {host}.")

        return


class Inventory(Command):
    def get_parser(self, prog_name):
        parser = super(Inventory, self).get_parser(prog_name)
        parser.add_argument(
            "host",
            nargs=1,
            type=str,
            help="Hostname (as the host is known in Ansible inventory)",
        )
        parser.add_argument(
            "variable",
            nargs="?",
            type=str,
            help="Name of a variable to show",
        )
        return parser

    def take_action(self, parsed_args):
        host = parsed_args.host[0]
        variable = parsed_args.variable

        try:
            result = subprocess.check_output(
                f"ansible-inventory -i /ansible/inventory/hosts.yml --host {host}",
                shell=True,
                stderr=subprocess.DEVNULL,
            )
        except subprocess.CalledProcessError:
            logger.error(f"Host {host} not found in inventory.")
            return

        data = json.loads(result)
        table = []

        if variable:
            if variable in data:
                row = pprint.pformat(data[variable], indent=2, width=60, compact=True)
                table.append([host, variable, row])
            else:
                logger.error(f"Variable {variable} not found in inventory for {host}.")
        else:
            for variable in data:
                row = pprint.pformat(data[variable], indent=2, width=60, compact=True)
                table.append([host, variable, row])

        if table:
            print(
                tabulate(table, headers=["Host", "Variable", "Value"], tablefmt="grid")
            )

        return



import sys

from cliff.app import App
from cliff.commandmanager import CommandManager

from . import __version__


class OsismApp(App):
    def __init__(self):
        super(OsismApp, self).__init__(
            description="OSISM manager interface",
            version=__version__,
            command_manager=CommandManager("osism.commands"),
            deferred_help=True,
        )


def main(argv=sys.argv[1:]):
    app = OsismApp()
    result = app.run(argv)
    return result


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



import subprocess
import time

from cliff.command import Command
from watchdog.observers.polling import PollingObserver
from watchdog.events import FileSystemEventHandler

from osism.tasks import reconciler


class Run(Command):
    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument("type", nargs=1, type=str, help="Type of the service")
        return parser

    def take_action(self, parsed_args):
        service = parsed_args.type[0]

        if service == "api":
            p = subprocess.Popen(
                "uvicorn osism.api:app --host 0.0.0.0 --port 8000", shell=True
            )
            p.wait()

        elif service == "listener":
            p = subprocess.Popen(
                "python3 -c 'from osism.services import listener; listener.main()'",
                shell=True,
            )
            p.wait()

        elif service == "beat":
            ts = [
                "osism.tasks.ansible",
                "osism.tasks.ceph",
                "osism.tasks.conductor",
                "osism.tasks.kolla",
                "osism.tasks.netbox",
                "osism.tasks.openstack",
                "osism.tasks.reconciler",
            ]
            ps = [
                subprocess.Popen(
                    f"celery -A {t} --broker=redis://redis beat -s /tmp/celerybeat-schedule-{t}.db",
                    shell=True,
                )
                for t in ts
            ]

            for p in ps:
                p.wait()

        elif service == "flower":
            p = subprocess.Popen("celery --broker=redis://redis flower", shell=True)
            p.wait()

        elif service == "reconciler":
            p = subprocess.Popen(
                "celery -A osism.tasks.reconciler worker -n reconciler --loglevel=INFO -Q reconciler",
                shell=True,
            )
            p.wait()

        elif service == "watchdog":
            event_handler_inventory = FileSystemEventHandler()
            event_handler_inventory.on_any_event = self.watchdog_inventory_on_any_event

            # We are not interested in being notified directly of any changes.
            # Therefore not the Inotify Observer but the Polling Observer. It
            # checks for changes every 10 seconds.

            observer = PollingObserver(10.0)
            observer.schedule(
                event_handler_inventory, "/opt/configuration/inventory", recursive=True
            )
            observer.start()

            try:
                while True:
                    time.sleep(1)
            finally:
                observer.stop()
                observer.join()

    def watchdog_inventory_on_any_event(self, event):
        reconciler.run.delay()



from celery import Celery

from osism.tasks import Config, run_ansible_in_environment

app = Celery("ansible")
app.config_from_object(Config)


@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    # 43200 seconds = 12 hours
    sender.add_periodic_task(43200.0, gather_facts.s(), expires=10)


@app.task(bind=True, name="osism.tasks.ansible.gather_facts")
def gather_facts(self, publish=True):
    return run_ansible_in_environment(self.request.id, "generic", "facts", [], publish)


@app.task(bind=True, name="osism.tasks.ansible.run")
def run(self, environment, playbook, arguments, publish=True):
    return run_ansible_in_environment(
        self.request.id, environment, playbook, arguments, publish
    )



import argparse
import time

from cliff.command import Command
from loguru import logger

from osism.core.enums import VALIDATE_PLAYBOOK2ENVIRONMENT
from osism.tasks import ansible
from osism.utils import redis


class Run(Command):
    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument(
            "validator",
            nargs=1,
            type=str,
            help="Validator to run",
            choices=VALIDATE_PLAYBOOK2ENVIRONMENT.keys(),
        )
        parser.add_argument(
            "arguments", nargs=argparse.REMAINDER, help="Other arguments for Ansible"
        )
        parser.add_argument(
            "--format",
            default="log",
            help="Output type",
            const="log",
            nargs="?",
            choices=["script", "log"],
        ),
        parser.add_argument(
            "--timeout",
            default=300,
            type=int,
            help="Timeout to end if there is no output",
        )
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the validator run has been completed",
            action="store_true",
        )
        return parser

    def _handle_task(self, t, wait, format, timeout):
        rc = 0
        if wait:
            p = redis.pubsub()
            p.subscribe(f"{t.task_id}")

            stoptime = time.time() + timeout
            while time.time() < stoptime:
                m = p.get_message(timeout=stoptime - time.time())
                if m:
                    stoptime = time.time() + timeout
                    if type(m["data"]) == bytes:
                        line = m["data"].decode("utf-8")
                        if line.startswith("RC: "):
                            rc = int(line[4:])
                            continue
                        if line == "QUIT":
                            redis.close()
                            # NOTE: Use better solution
                            return rc
                        print(line, end="")
                else:
                    logger.info(
                        f"No further output after {timeout} seconds. Therefore finish."
                    )
                    return rc

        else:
            if format == "log":
                logger.info(
                    f"Task {t.task_id} is running in background. No more output. Check ARA for logs."
                )
            elif format == "script":
                print(f"{t.task_id}")

            return rc

    def take_action(self, parsed_args):
        arguments = parsed_args.arguments
        validator = parsed_args.validator[0]
        format = parsed_args.format
        timeout = parsed_args.timeout
        wait = not parsed_args.no_wait

        environment = VALIDATE_PLAYBOOK2ENVIRONMENT[validator]
        t = ansible.run.delay(environment, f"validate-{validator}", arguments)
        rc = self._handle_task(t, wait, format, timeout)

        return rc



import os
import time

from kombu import Connection, Exchange, Queue
from kombu.mixins import ConsumerMixin
from loguru import logger
import json

from osism import utils
from osism.tasks import netbox, openstack

EXCHANGE_NAME = "ironic"
ROUTING_KEY = "ironic_versioned_notifications.info"
QUEUE_NAME = "osism-listener-ironic"
BROKER_URI = os.getenv("BROKER_URI")


class NotificationsDump(ConsumerMixin):
    def __init__(self, connection):
        self.connection = connection
        return

    def get_consumers(self, consumer, channel):
        exchange = Exchange(EXCHANGE_NAME, type="topic", durable=False)
        queue = Queue(
            QUEUE_NAME,
            exchange,
            routing_key=ROUTING_KEY,
            durable=False,
            auto_delete=True,
            no_ack=True,
        )
        return [consumer(queue, callbacks=[self.on_message])]

    def on_message(self, body, message):
        data = json.loads(body["oslo.message"])
        # logger.info(data)

        event_type = data["event_type"]
        payload = data["payload"]
        object_data = payload["ironic_object.data"]

        name = object_data["name"]

        # References:
        #
        # * https://docs.openstack.org/ironic/latest/admin/notifications.html#available-notifications
        # * https://docs.openstack.org/ironic/latest/user/states.html
        # * https://docs.openstack.org/ironic/latest/_images/states.svg

        if event_type == "baremetal.node.power_set.end":
            logger.info(
                f"baremetal.node.power_set.end ## {name} ## {object_data['power_state']}"
            )
            netbox.set_state.delay(name, object_data["power_state"], "power")

        elif event_type == "baremetal.node.power_state_corrected.success":
            logger.info(
                f"baremetal.node.power_state_corrected.success ## {name} ## {object_data['power_state']}"
            )
            netbox.set_state.delay(name, object_data["power_state"], "power")

        elif event_type == "baremetal.node.maintenance_set.end":
            logger.info(
                f"baremetal.node.maintenance_set.end ## {name} ## {object_data['maintenance']}"
            )
            netbox.set_maintenance.delay(name, object_data["maintenance"])

        elif event_type == "baremetal.node.provision_set.start":
            logger.info(
                f"baremetal.node.provision_set.start ## {name} ## {object_data['provision_state']}"
            )

            if object_data["event"] == "inspect":
                # system should be in state a
                netbox.connect.delay(name, "a")

            if object_data["provision_state"] == "cleaning":
                # system should be in state b
                netbox.connect.delay(name, "b")

            if object_data["provision_state"] == "available":
                # system should be in state c
                netbox.connect.delay(name, "c")

            if object_data["target_provision_state"] == "active":
                pass

        # A provision status was successfully set, update it in the netbox
        elif event_type == "baremetal.node.provision_set.success":
            logger.info(
                f"baremetal.node.provision_set.success ## {name} ## {object_data['provision_state']}"
            )
            netbox.set_state.delay(name, object_data["provision_state"], "provision")

            if object_data["provision_state"] == "manageable":

                # system should be in state c
                netbox.connect.delay(name, "c")

        elif event_type == "baremetal.node.provision_set.end":
            logger.info(
                f"baremetal.node.provision_set.end ## {name} ## {object_data['provision_state']}"
            )
            netbox.set_state.delay(name, object_data["provision_state"], "provision")

            if (
                object_data["previous_provision_state"] == "inspect wait"
                and object_data["event"] == "done"
            ):
                netbox.set_state.delay(name, "introspected", "introspection")
                openstack.baremetal_set_node_provision_state.delay(name, "provide")

            elif object_data["previous_provision_state"] == "wait call-back":
                pass

            elif (
                object_data["previous_provision_state"] == "cleaning"
                and object_data["provision_state"] == "available"
            ):  # noqa

                # system should be in state c
                netbox.connect.delay(name, "c")

        elif event_type == "baremetal.port.create.end":
            logger.info(f"baremetal.port.create.end ## {object_data['uuid']}")

            mac_address = object_data["address"]
            interface_a = utils.nb.dcim.interfaces.get(mac_address=mac_address)
            device_a = interface_a.device

            task = openstack.baremetal_get_network_interface_name.delay(
                device_a.name, mac_address
            )
            task.wait(timeout=None, interval=0.5)
            network_interface_name = task.get()

            netbox.update_network_interface_name.delay(
                object_data["address"], network_interface_name
            )

        elif event_type == "baremetal.port.update.end":
            logger.info(f"baremetal.port.update.end ## {object_data['uuid']}")

            mac_address = object_data["address"]
            interface_a = utils.nb.dcim.interfaces.get(mac_address=mac_address)
            device_a = interface_a.device

            task = openstack.baremetal_get_network_interface_name.delay(
                device_a.name, mac_address
            )
            task.wait(timeout=None, interval=0.5)
            network_interface_name = task.get()

            netbox.update_network_interface_name.delay(
                object_data["address"], network_interface_name
            )

        elif event_type == "baremetal.node.delete.end":
            logger.info(f"baremetal.node.delete.end ## {name}")

            netbox.set_state.delay(name, "unregistered", "ironic")
            netbox.set_state.delay(name, None, "provision")
            netbox.set_state.delay(name, None, "power")
            netbox.set_state.delay(name, None, "introspection")
            netbox.set_state.delay(name, None, "deployment")

            # system should be in state a
            netbox.connect.delay(name, "a")

            # remove internal flavor
            openstack.baremetal_delete_internal_flavor.delay(name)

        else:
            logger.info(f"{event_type} ## {name}")

        logger.info(object_data)


def main():
    try:
        with Connection(BROKER_URI, connect_timeout=30.0) as connection:
            connection.connect()
            NotificationsDump(connection).run()
    except ConnectionRefusedError:
        logger.error("Connection with broker refused. Retry in 60 seconds.")
        time.sleep(60)


if __name__ == "__main__":
    main()



import os
import subprocess

from celery import Celery
from celery.signals import worker_process_init
import json
import pynetbox
from redis import Redis

from osism import settings
from osism.actions import (
    check_configuration,
    deploy_configuration,
    diff_configuration,
    generate_configuration,
    manage_device,
    manage_interface,
)
from osism.tasks import Config, ansible, openstack

app = Celery("netbox")
app.config_from_object(Config)

redis = None
nb = None


@worker_process_init.connect
def celery_init_worker(**kwargs):
    global nb
    global redis

    redis = Redis(host="redis", port="6379")
    nb = pynetbox.api(settings.NETBOX_URL, token=settings.NETBOX_TOKEN)

    if settings.IGNORE_SSL_ERRORS:
        import requests

        requests.packages.urllib3.disable_warnings()
        session = requests.Session()
        session.verify = False
        nb.http_session = session


@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):

    if Config.enable_bifrost in ["True", "true", "Yes", "yes"]:
        # Synchronize the status of Bifrost with Netbox every 5 minutes
        sender.add_periodic_task(300.0, periodic_synchronize_bifrost.s(), expires=10)


@app.task(bind=True, name="osism.tasks.netbox.periodic_synchronize_ironic")
def periodic_synchronize_ironic(self):
    """Synchronize the state of Ironic with Netbox"""
    openstack.baremetal_node_list.apply_async((), link=synchronize_device_state.s())


@app.task(bind=True, name="osism.tasks.netbox.periodic_synchronize_bifrost")
def periodic_synchronize_bifrost(self):
    """Synchronize the state of Bifrost with Netbox"""
    ansible.run.apply_async(
        ("manager", "bifrost-command", "baremetal node list -f json"),
        link=synchronize_device_state.s(),
    )


@app.task(bind=True, name="osism.tasks.netbox.run")
def run(self, action, arguments):
    pass


@app.task(bind=True, name="osism.tasks.netbox.update_network_interface_name")
def update_network_interface_name(self, mac_address, network_interface_name):
    manage_interface.update_network_interface_name(mac_address, network_interface_name)


@app.task(bind=True, name="osism.tasks.netbox.import_device_types")
def import_device_types(self, vendors, library=False):
    global redis

    if library:
        env = {**os.environ, "BASE_PATH": "/devicetype-library/device-types/"}
    else:
        env = {**os.environ, "BASE_PATH": "/netbox/device-types/"}

    if vendors:
        p = subprocess.Popen(
            f"python3 /import/main.py --vendors {vendors}",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=env,
        )
    else:
        p = subprocess.Popen(
            "python3 /import/main.py",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=env,
        )

    p.communicate()


@app.task(bind=True, name="osism.tasks.netbox.synchronize_device_state")
def synchronize_device_state(self, data):
    """Synchronize the state of Bifrost or Ironic with Netbox"""

    if type(data) == str:
        data = json.loads(data)

    if not data:
        return

    for device in data:
        manage_device.set_provision_state(device["Name"], device["Provisioning State"])
        manage_device.set_power_state(device["Name"], device["Power State"])


@app.task(bind=True, name="osism.tasks.netbox.states")
def states(self, data):
    result = manage_device.get_states(data.keys())
    return result


@app.task(bind=True, name="osism.tasks.netbox.transitions")
def transitions(self, data):
    result = manage_device.get_transitions(data.keys())
    return result


@app.task(bind=True, name="osism.tasks.netbox.data")
def data(self, collection, device, state):
    result = manage_device.load_data_from_filesystem(collection, device, state)
    return result


@app.task(bind=True, name="osism.tasks.netbox.connect")
def connect(self, device=None, state=None, data={}, enforce=False):
    manage_device.run(device, state, data, enforce)


@app.task(bind=True, name="osism.tasks.netbox.set_state")
def set_state(self, device=None, state=None, state_type=None):
    manage_device.set_state(device, state, state_type)


@app.task(bind=True, name="osism.tasks.netbox.set_maintenance")
def set_maintenance(self, device=None, state=None):
    manage_device.set_maintenance(device, state)


@app.task(bind=True, name="osism.tasks.netbox.disable")
def disable(self, name):
    global nb

    for interface in nb.dcim.interfaces.filter(device=name):
        if str(interface.type) in ["Virtual"]:
            continue

        if "Port-Channel" in interface.name:
            continue

        if not interface.connected_endpoint and interface.enabled:
            interface.enabled = False
            interface.save()

        # FIXME: only enable devices that are not disabled by configuration
        if interface.connected_endpoint and not interface.enabled:
            interface.enabled = True
            interface.save()


@app.task(bind=True, name="osism.tasks.netbox.generate")
def generate(self, name, template=None):
    generate_configuration.for_device(name, template)


@app.task(bind=True, name="osism.tasks.netbox.deploy")
def deploy(self, name):
    deploy_configuration.for_device(name)


@app.task(bind=True, name="osism.tasks.netbox.check")
def check(self, name):
    check_configuration.for_device(name)


@app.task(bind=True, name="osism.tasks.netbox.diff")
def diff(self, name):
    diff_configuration.for_device(name)


@app.task(bind=True, name="osism.tasks.netbox.get_devices_not_yet_registered_in_ironic")
def get_devices_not_yet_registered_in_ironic(
    self, status="active", tags=["managed-by-ironic"], ironic_enabled=True
):
    global nb

    devices = nb.dcim.devices.filter(
        tag=tags, status=status, cf_ironic_enabled=[ironic_enabled]
    )

    result = []

    for device in devices:
        if (
            "ironic_state" in device.custom_fields
            and device.custom_fields["ironic_state"] != "registered"
        ):
            result.append(device.name)

    return result


@app.task(
    bind=True,
    name="osism.tasks.netbox.get_devices_that_should_have_an_allocation_in_ironic",
)
def get_devices_that_should_have_an_allocation_in_ironic(self):
    global nb

    devices = nb.dcim.devices.filter(
        tag=["managed-by-ironic", "managed-by-osism"],
        status="active",
        cf_ironic_enabled=[True],
        cf_ironic_state=["registered"],
        cf_provision_state=["available"],
        cf_introspection_state=["introspected"],
        cf_device_type=["server"],
    )

    result = []

    for device in devices:
        result.append(device.name)

    return result


@app.task(bind=True, name="osism.tasks.netbox.ping")
def ping(self):
    global nb

    status = nb.status()

    return status



import argparse
import subprocess

from cliff.command import Command


class Ansible(Command):
    def get_parser(self, prog_name):
        parser = super(Ansible, self).get_parser(prog_name)
        parser.add_argument(
            "parameter",
            nargs=argparse.REMAINDER,
            type=str,
            help="Parameters to add (all paraemters of the ara command are possible)",
        )
        return parser

    def take_action(self, parsed_args):
        parameters = " ".join(parsed_args.parameter)
        subprocess.call(
            f"/usr/local/bin/ara {parameters}",
            shell=True,
        )


class Container(Command):
    def get_parser(self, prog_name):
        parser = super(Container, self).get_parser(prog_name)
        parser.add_argument("host", nargs=1, type=str, help="Hostname or address")
        parser.add_argument(
            "container", nargs=1, type=str, help="Name of the container"
        )
        parser.add_argument(
            "parameter",
            nargs=argparse.REMAINDER,
            type=str,
            help="Parameters to add (all paraemters of the docker logs command are possible)",
        )
        return parser

    def take_action(self, parsed_args):
        host = parsed_args.host[0]
        container_name = parsed_args.container[0]
        parameters = " ".join(parsed_args.parameter)

        ssh_command = f"docker logs {parameters} {container_name}"
        ssh_options = "-o StrictHostKeyChecking=no -o LogLevel=ERROR"

        # FIXME: use paramiko or something else more Pythonic + make operator user + key configurable
        subprocess.call(
            f"/usr/bin/ssh -i /ansible/secrets/id_rsa.operator {ssh_options} dragon@{host} {ssh_command}",
            shell=True,
        )


class File(Command):
    def get_parser(self, prog_name):
        parser = super(File, self).get_parser(prog_name)
        parser.add_argument("host", nargs=1, type=str, help="Hostname or address")
        return parser

    def take_action(self, parsed_args):
        print("NOT YET IMPLEMENTED")



from datetime import datetime
import ipaddress
import os

from loguru import logger
import git
import gitdb
import jinja2
from pottery import Redlock

from osism.utils import first
from osism import utils


def vlans_as_string(untagged_vlan, tagged_vlans):
    vlans = [str(x.vid) for x in tagged_vlans]
    if untagged_vlan:
        vlans = vlans + [str(untagged_vlan.vid)]
    return ",".join(sorted(vlans))


def for_device(name, template=None):
    device = utils.nb.dcim.devices.get(name=name)

    if (
        "device_type" not in device.custom_fields
        or device.custom_fields["device_type"] != "switch"
    ):
        return

    if "Managed by OSISM" not in [str(x) for x in device.tags]:
        return

    logger.info(f"Generate configuration for device {device.name}")

    if not template:
        template = f"{device.name}.cfg.j2"

        if not os.path.isfile(f"/netbox/templates/{template}"):
            template = f"{device.device_type.manufacturer.name}.cfg.j2"

        if not os.path.isfile(f"/netbox/templates/{template}"):
            template = "default.cfg.j2"

    vlans = utils.nb.ipam.vlans.filter(available_on_device=device.id)
    interfaces = utils.nb.dcim.interfaces.filter(device=device)

    interfaces_ethernet = []
    interfaces_port_channels = []
    interfaces_virtual = []

    for interface in interfaces:
        if str(interface.type) == "Link Aggregation Group (LAG)":
            interfaces_port_channels.append(interface)
        elif str(interface.type) == "Virtual":
            interfaces_virtual.append(interface)
        else:
            interfaces_ethernet.append(interface)

    # Port-Channel10 + Vlan4094 are always used as MLAG
    try:
        mlag = utils.nb.dcim.interfaces.get(device=device, name="Port-Channel10")
        mlag_vlan = utils.nb.dcim.interfaces.get(device=device, name="Vlan4094")
        mlag_address = utils.nb.ipam.ip_addresses.get(
            device=device, interface="Vlan4094"
        )
    except:  # noqa
        mlag = None
        mlag_vlan = None
        mlag_address = None

    # NOTE: only work with /30
    try:
        x = list(ipaddress.ip_interface(mlag_address.address).network.hosts())
        x.remove(ipaddress.ip_interface(mlag_address.address).ip)
        mlag_peer_address = x[0]
    except:  # noqa
        mlag_peer_address = None

    try:
        mlag_domain_id = device.name.split("-")[1]
    except:  # noqa
        mlag_domain_id = None

    repo = git.Repo.init(path="/state")
    repo.config_writer().set_value("user", "name", "Netbox Generator").release()
    repo.config_writer().set_value(
        "user", "email", "netbox-generator@reconciler.local"
    ).release()

    data = {
        "hostname": device.name,
        "interfaces_ethernet": interfaces_ethernet,
        "interfaces_virtual": interfaces_virtual,
        "interfaces_port_channels": interfaces_port_channels,
        "vlans": vlans,
        "device": name,
        "nb": utils.nb,
        "first": first,
        "vlans_as_string": vlans_as_string,
        "mlag": mlag,
        "mlag_vlan": mlag_vlan,
        "mlag_peer_address": mlag_peer_address,
        "mlag_domain_id": mlag_domain_id,
        "timestamp": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        "device_type": device.device_type.model,
        "device_manufacturer": device.device_type.manufacturer.name,
    }

    loader = jinja2.FileSystemLoader(searchpath="/netbox/templates/")
    environment = jinja2.Environment(loader=loader)
    template = environment.get_template(template)
    result = template.render(data)

    logger.info(f"Writing generated configuration to /state/{device.name}.cfg.j2")
    with open(f"/state/{device.name}.cfg.j2", "w+") as fp:
        fp.write(os.linesep.join([s for s in result.splitlines() if s]))

    # Allow only one change per time
    lock = Redlock(key="lock_repository", masters={utils.redis})
    lock.acquire()

    repo.git.add(f"/state/{device.name}.cfg.j2")

    try:
        if len(repo.index.diff("HEAD")) > 0:
            logger.info(f"Committing changes in /state/{device.name}.cfg.j2")
            repo.git.commit(message=f"Update {device.name}")

    # Ref 'HEAD' did not resolve to an object
    except gitdb.exc.BadName:
        logger.info("Initial commit")
        repo.git.commit(message="Initial commit")

    lock.release()



from celery import Celery

from osism.tasks import Config, run_ansible_in_environment

app = Celery("ceph")
app.config_from_object(Config)


@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    pass


@app.task(bind=True, name="osism.tasks.ceph.run")
def run(self, playbook, arguments, publish=True):
    return run_ansible_in_environment(
        self.request.id, "ceph", playbook, arguments, publish
    )



import subprocess

from cliff.command import Command
from loguru import logger

from osism.tasks import reconciler


class Run(Command):
    def take_action(self, parsed_args):
        logger.info(
            "The osism reconciler command is deprecated and will be removed. Use osism service reconciler."
        )
        # NOTE: use python interface in the future, works for the moment
        p = subprocess.Popen(
            "celery -A osism.tasks.reconciler worker -n reconciler --loglevel=INFO -Q reconciler",
            shell=True,
        )
        p.wait()


class Sync(Command):
    def get_parser(self, prog_name):
        parser = super(Sync, self).get_parser(prog_name)
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the sync has been completed",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        wait = not parsed_args.no_wait

        task = reconciler.run.delay()
        if wait:
            logger.info(f"Task {task.task_id} is running. Wait. No more output.")
            task.wait(timeout=None, interval=0.5)



import argparse
import time

from celery import group
from celery.result import GroupResult
from cliff.command import Command
from loguru import logger

from osism.core import enums
from osism.tasks import ansible, ceph, kolla
from osism.utils import redis


class Run(Command):
    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument(
            "--environment", type=str, help="Environment that is to be used explicitly"
        )
        parser.add_argument("role", nargs=1, type=str, help="Role to be applied")
        parser.add_argument(
            "arguments", nargs=argparse.REMAINDER, help="Other arguments for Ansible"
        )
        parser.add_argument(
            "--format",
            default="log",
            help="Output type",
            const="log",
            nargs="?",
            choices=["script", "log"],
        ),
        parser.add_argument(
            "--timeout",
            default=300,
            type=int,
            help="Timeout to end if there is no output",
        )
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the role has been applied",
            action="store_true",
        )
        return parser

    def _handle_loadbalancer(self, t, wait, format, timeout):
        # process the parent task
        rc = self._handle_task(t.parent, wait, format, timeout)

        # It is necessary to wait for all task even if this is not excpected by the
        # user because of the following exception thrown by the garbage collector.
        #
        # Exception ignored in: <function AsyncResult.__del__ at 0x7f8c91ac74c0>
        # Traceback (most recent call last):
        # [...]
        # ImportError: sys.meta_path is None, Python is likely shutting down

        if not wait:
            t.parent.get()

        # process the child tasks
        if format == "log":
            for c in t.children:
                logger.info(
                    f"Task {c.task_id} is running in background. No more output. Check ARA for logs."
                )

        # As explained above, it is neceesary to wait for all tasks.
        t.get()

        return rc

    def _handle_task(self, t, wait, format, timeout):
        rc = 0
        if wait:
            p = redis.pubsub()
            p.subscribe(f"{t.task_id}")

            stoptime = time.time() + timeout
            while time.time() < stoptime:
                m = p.get_message(timeout=stoptime - time.time())
                if m:
                    stoptime = time.time() + timeout
                    if type(m["data"]) == bytes:
                        line = m["data"].decode("utf-8")
                        if line.startswith("RC: "):
                            rc = int(line[4:])
                            continue
                        if line == "QUIT":
                            redis.close()
                            # NOTE: Use better solution
                            return rc
                        print(line, end="")
                else:
                    logger.info(
                        f"No further output after {timeout} seconds. Therefore finish."
                    )
                    return rc

        else:
            if format == "log":
                logger.info(
                    f"Task {t.task_id} is running in background. No more output. Check ARA for logs."
                )
            elif format == "script":
                print(f"{t.task_id}")

            return rc

    def handle_role(self, arguments, environment, role, wait, format, timeout):
        if not environment:
            try:
                environment = enums.MAP_ROLE2ENVIRONMENT[role]
            except:  # noqa: E722
                environment = "custom"

        if environment == "ceph":
            if role.startswith("ceph-"):
                t = ceph.run.delay(role[5:], arguments)
            else:
                t = ceph.run.delay(role, arguments)
        elif environment == "kolla":
            if role.startswith("kolla-"):
                t = kolla.run.delay(role[6:], arguments)
            else:
                t = kolla.run.delay(role, arguments)
        elif role == "loadbalancer-ng":
            g = group(
                kolla.run.si(playbook, arguments)
                for playbook in enums.LOADBALANCER_PLAYBOOKS
            )
            t = (kolla.run.s("loadbalancer-ng", arguments) | g).apply_async()
        else:
            t = ansible.run.delay(environment, role, arguments)

        if isinstance(t, GroupResult):
            rc = self._handle_loadbalancer(t, wait, format, timeout)
        else:
            rc = self._handle_task(t, wait, format, timeout)

        return rc

    def take_action(self, parsed_args):
        arguments = parsed_args.arguments
        environment = parsed_args.environment
        format = parsed_args.format
        role = parsed_args.role[0]
        timeout = parsed_args.timeout
        wait = not parsed_args.no_wait

        if role in enums.MAP_ROLE2ROLE:
            for r in enums.MAP_ROLE2ROLE[role]:
                rc = self.handle_role(arguments, environment, r, wait, format, timeout)
                if rc != 0:
                    break
        else:
            rc = self.handle_role(arguments, environment, role, wait, format, timeout)

        return rc



import argparse

from cliff.command import Command
from redis import Redis

from osism.tasks import ansible


redis = Redis(host="redis", port="6379")


class Run(Command):
    def get_parser(self, prog_name):
        parser = super(Run, self).get_parser(prog_name)
        parser.add_argument(
            "arguments", nargs=argparse.REMAINDER, help="Arguments for Bifrost"
        )
        return parser

    def take_action(self, parsed_args):
        task = ansible.run.delay("manager", "bifrost-command", parsed_args.arguments)

        task.wait(timeout=None, interval=0.5)

        result = task.get()
        print(result)


class Deploy(Command):
    def get_parser(self, prog_name):
        parser = super(Deploy, self).get_parser(prog_name)
        parser.add_argument(
            "--no-wait",
            default=False,
            help="Do not wait until the role has been applied",
            action="store_true",
        )
        return parser

    def take_action(self, parsed_args):
        wait = not parsed_args.no_wait

        ansible.run.delay("manager", "bifrost-deploy", [])

        if wait:
            p = redis.pubsub()

            # NOTE: use task_id or request_id in future
            p.subscribe("manager-bifrost-deploy")

            while True:
                for m in p.listen():
                    if type(m["data"]) == bytes:
                        if m["data"].decode("utf-8") == "QUIT":
                            redis.close()
                            # NOTE: Use better solution
                            return
                        print(m["data"].decode("utf-8"), end="")
